package wacc

import scala.collection.mutable

object typeChecker {
  import wacc.ast._
  import wacc.types._

  case class TypeCheckerCtx(
      globalNames: Map[Ident, SemType],
      errors: mutable.Builder[Error, List[Error]]
  ) {
    def typeOf(ident: Ident): SemType = globalNames.withDefault { case Ident(_, -1) => ? }(ident)

    def error(err: Error): SemType =
      errors += err
      ?
  }

  enum Constraint {
    case Unconstrained
    case Is(ty: SemType, msg: String)
    case IsSymmetricCompatible(ty: SemType, msg: String)
    case IsUnweakanable(ty: SemType, msg: String)
    case IsVar(msg: String)
    case IsEither(ty1: SemType, ty2: SemType, msg: String)
    case Never(msg: String)
  }

  extension (ty: SemType)
    infix def satisfies(constraint: Constraint): SemType = (ty, constraint) match {
      case (KnownType.Array(KnownType.Char), Constraint.Is(KnownType.String, _)) =>
        KnownType.String
      case (
            KnownType.String,
            Constraint.IsSymmetricCompatible(KnownType.Array(KnownType.Char), _)
          ) =>
        KnownType.String
      case (ty, Constraint.IsSymmetricCompatible(ty2, msg)) => ty satisfies Constraint.Is(ty2, msg)
      case (ty, Constraint.Is(ty2, msg)) => ty satisfies Constraint.IsUnweakanable(ty2, msg)
    }

  def check(prog: Program)(using
      ctx: TypeCheckerCtx
  ): Unit = {
    prog.funcs.foreach { case FuncDecl(_, name, _, stmts) =>
      val retType = ctx.typeOf(name)
      stmts.toList.foreach(
        checkStmt(_, Constraint.Is(retType, s"function ${name.v} must return $retType"))
      )
    }
    prog.main.toList.foreach(checkStmt(_, Constraint.Never("main function must not return")))
  }

  private def checkStmt(stmt: Stmt, returnConstraint: Constraint)(using
      ctx: TypeCheckerCtx
  ): Unit = stmt match {
    case VarDecl(_, name, value) =>
      val expectedTy = ctx.typeOf(name)
      checkValue(
        value,
        Constraint.Is(
          expectedTy,
          s"variable ${name.v} must be assigned a value of type $expectedTy"
        )
      )
    case Assign(lhs, rhs) =>
      val lhsTy = checkValue(lhs, Constraint.Unconstrained)
      checkValue(rhs, Constraint.Is(lhsTy, s"assignment must have type $lhsTy"))
    case Read(lhs) =>
      checkValue(
        lhs,
        Constraint.IsEither(KnownType.Int, KnownType.Char, "read must be int or char")
      )
    case Free(lhs) =>
      checkValue(
        lhs,
        Constraint.IsEither(
          KnownType.Array(?),
          KnownType.Pair(?, ?),
          "free must be an array or pair"
        )
      )
    case Return(expr) =>
      checkValue(expr, returnConstraint)
    case Exit(expr) =>
      checkValue(expr, Constraint.Is(KnownType.Int, "exit value must be int"))
    case Print(expr, _) =>
      // This constraint should never fail, the scope-checker should have caught it already
      checkValue(expr, Constraint.IsVar("print value must be a variable"))
    case If(cond, thenStmt, elseStmt) =>
      checkValue(cond, Constraint.Is(KnownType.Bool, "if condition must be a bool"))
      thenStmt.toList.foreach(checkStmt(_, returnConstraint))
      elseStmt.toList.foreach(checkStmt(_, returnConstraint))
    case While(cond, body) =>
      checkValue(cond, Constraint.Is(KnownType.Bool, "while condition must be a bool"))
      body.toList.foreach(checkStmt(_, returnConstraint))
    case Block(body) =>
      body.toList.foreach(checkStmt(_, returnConstraint))
    case Skip() => ()
  }

  private def checkValue(value: LValue | RValue | Expr, constraint: Constraint)(using
      ctx: TypeCheckerCtx
  ): SemType = value match {
    case IntLiter(_)  => KnownType.Int satisfies constraint
    case BoolLiter(_) => KnownType.Bool satisfies constraint
    case CharLiter(_) => KnownType.Char satisfies constraint
    case StrLiter(_)  => KnownType.String satisfies constraint
    case PairLiter()  => KnownType.Pair(?, ?) satisfies constraint
    case id: Ident =>
      ctx.typeOf(id) satisfies constraint
    case ArrayElem(id, indices) =>
      val arrayTy = ctx.typeOf(id)
      val elemTy = indices.toList.foldRight(arrayTy) { (elem, acc) =>
        checkValue(elem, Constraint.Is(KnownType.Int, "array index must be an int"))
        acc match {
          case KnownType.Array(innerTy) => innerTy
          case _ =>
            ctx.error(
              Error.TypeMismatch(elem.pos, KnownType.Array(?), acc, "cannot index into a non-array")
            )
        }
      }
      elemTy satisfies constraint
    case Parens(expr) => checkValue(expr, constraint)
    case ArrayLiter(elems) =>
      KnownType.Array(elems.foldRight[SemType](?) { case (elem, acc) =>
        checkValue(
          elem,
          Constraint.IsSymmetricCompatible(acc, "array elements must have the same type")
        )
      }) satisfies constraint
    case NewPair(fst, snd) =>
      KnownType.Pair(
        checkValue(fst, Constraint.Unconstrained),
        checkValue(snd, Constraint.Unconstrained)
      ) satisfies constraint
    case Call(id, args) =>
      val funcTy = ctx.typeOf(id)
      funcTy match {
        case KnownType.Func(retTy, paramTys) => // TODO do we check argument lengths match
          args.zip(paramTys).foreach { case (arg, paramTy) =>
            checkValue(arg, Constraint.Is(paramTy, s"argument type mismatch in function ${id.v}"))
          }
          retTy satisfies constraint
        // Should never happen, the scope-checker should have caught this already
        // ctx error had it not
        case _ => ctx.error(Error.InternalError(id.pos, "function call to non-function"))
      }
    case Fst(elem) =>
      checkValue(
        elem,
        Constraint.Is(KnownType.Pair(?, ?), "fst must be applied to a pair")
      ) match {
        case KnownType.Pair(left, _) => left satisfies constraint
        case ?                       => ? satisfies constraint
        case _ => ctx.error(Error.InternalError(elem.pos, "fst must be applied to a pair"))
      } // satisfies constraint
    case Snd(elem) =>
      checkValue(
        elem,
        Constraint.Is(KnownType.Pair(?, ?), "snd must be applied to a pair")
      ) match {
        case KnownType.Pair(_, right) => right satisfies constraint
        case ?                        => ? satisfies constraint
        case _ => ctx.error(Error.InternalError(elem.pos, "snd must be applied to a pair"))
      }

    // Unary operators
    case Negate(x) =>
      checkValue(x, Constraint.Is(KnownType.Int, "negation must be applied to an int"))
      KnownType.Int satisfies constraint
    case Not(x) =>
      checkValue(x, Constraint.Is(KnownType.Bool, "logical not must be applied to a bool"))
      KnownType.Bool satisfies constraint
    case Len(x) =>
      checkValue(x, Constraint.Is(KnownType.Array(?), "len must be applied to an array"))
      KnownType.Int satisfies constraint
    case Ord(x) =>
      checkValue(x, Constraint.Is(KnownType.Char, "ord must be applied to a char"))
      KnownType.Int satisfies constraint
    case Chr(x) =>
      checkValue(x, Constraint.Is(KnownType.Int, "chr must be applied to an int"))
      KnownType.Char satisfies constraint

    // Binary operators
    case op: (Add | Sub | Mul | Div | Mod) =>
      val operand = Constraint.Is(KnownType.Int, "binary operator must be applied to an int")
      checkValue(op.x, operand)
      checkValue(op.y, operand)
      KnownType.Int satisfies constraint
    case op: (Eq | Neq) =>
      val xTy = checkValue(op.x, Constraint.Unconstrained)
      checkValue(op.y, Constraint.Is(xTy, "equality must be applied to values of the same type"))
      KnownType.Bool satisfies constraint
    case op: (Less | LessEq | Greater | GreaterEq) =>
      val xTy = checkValue(
        op.x,
        Constraint.IsEither(
          KnownType.Int,
          KnownType.Char,
          "comparison must be applied to an int or char"
        )
      )
      checkValue(op.y, Constraint.Is(xTy, "comparison must be applied to values of the same type"))
      KnownType.Bool satisfies constraint
    case op: (And | Or) =>
      val operand = Constraint.Is(KnownType.Bool, "logical operator must be applied to a bool")
      checkValue(op.x, operand)
      checkValue(op.y, operand)
      KnownType.Bool satisfies constraint
  }
}
